/*
 * Copyright 2002 Draagon Software LLC. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software LLC.
 * Use is subject to license terms.
 */

package com.draagon.meta;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.util.ArrayList;
import java.util.Collection;

public abstract class MetaClass extends MetaData
{
  private static Log log = LogFactory.getLog( MetaClass.class );

  /** Object class name attribute */
  public final static String ATTR_OBJECT = "object";

  /** Separator for package and class names */
  public final static String SEPARATOR = "::";

  /**
   * Constructs the MetaClass
   */
  public MetaClass()
  {
  }

  /**
   * Gets the primary MetaData class
   */
  public final Class<?> getMetaDataClass()
  {
    return MetaClass.class;
  }

  /**
   * Returns the MetaClass for the specified Meta Class name
   */
  public static MetaClass forName( String name )
    //throws MetaClassNotFoundException
  {
    return MetaClassLoader.findMetaClassWithName( name );
  }

  /**
   * Sets the Package for the MetaClass
   */
  /*public void setPackage( String packageName )
    //throws MetaException
  {
    mPackage = packageName;
  }*/

  /**
   * Retrieve the MetaClass package
   */
  public String getPackage()
  {
    String name = getName();
    if ( name == null ) return null;
    int i = name.lastIndexOf( SEPARATOR );
    if ( i >= 0 ) return name.substring( 0, i );
    return "";
  }

  /**
   * Retrieve the MetaClass package
   */
  public String getShortName()
  {
    String name = getName();
    if ( name == null ) return null;
    int i = name.lastIndexOf( SEPARATOR );
    if ( i >= 0 ) return name.substring( i + SEPARATOR.length() );
    return name;
  }

  /**
   * Returns the MetaClassLoader
   */
  public MetaClassLoader getClassLoader()
  {
    return (MetaClassLoader) getParent();
  }

  /**
   * Sets the Super Class
   */
  public void setSuperClass( MetaClass superClass )
  {
    setSuperData( superClass );

    // for each field, create a new one and attach the super field
  }

  /**
   * Gets the Super Class
   */
  public MetaClass getSuperClass()
  {
    return (MetaClass) getSuperData();
  }

  /**
   * Retrieve the name of the MetaClass
   */
  /*public String getFullName()
  {
    if ( getPackage() != null && getPackage().length() > 0 )
    {
      return getPackage() + SEPARATOR + getName();
    }
    else
      return getName();
  }*/

  ////////////////////////////////////////////////////
  // FIELD METHODS

  /**
   * Return the MetaField count
   */
  public Collection<MetaField> getMetaFields()
  {
	Collection<MetaData> children = getChildren( MetaField.class );
	Collection<MetaField> fields = new ArrayList<MetaField>( children.size() );
	for ( MetaData md : children ) fields.add( (MetaField) md );
	return fields;
  }

  /**
   * Add a field to the MetaClass
   */
  public void addMetaField( MetaField f )
    //throws InvalidMetaDataException
  {
    addChild( f );
  }

  /**
   * Whether the named MetaField exists
   */
  public boolean hasMetaField( String name )
  {
    try {
      getMetaField( name );
      return true;
    }
    catch( MetaFieldNotFoundException e ) {
      return false;
    }
  }

  /**
   * Return the specified MetaField of the MetaClass
   */
  public MetaField getMetaField( String name )
      //throws MetaFieldNotFoundException
  {
	final String KEY = "getMetaField(" + name + ")";

	MetaField f = (MetaField) getCacheValue( KEY );

	if ( f == null )
	{
	    try {
	      f = (MetaField) getChild( name, MetaField.class );
	    }
	    catch( MetaDataNotFoundException e )
	    {
	      if ( getSuperClass() != null )
	      {
	        try {
	          f = getSuperClass().getMetaField( name );
	        } catch( MetaFieldNotFoundException ex ) {}
	      }

	      throw new MetaFieldNotFoundException( "MetaField [" + name + "] does not exist in MetaClass [" + toString() + "]" );
	    }

	    setCacheValue( KEY, f );
	}

	return f;
  }

  /**
   * Whether the object is aware of its own state.  This is false
   *  by default, which means all state methods will throw an
   *  UnsupportedOperationException.
   */
  public boolean isStateAware() {
    return false;
  }

  ////////////////////////////////////////////////////
  // OBJECT METHODS

  /**
   * Retrieves the object class of an object, or null if one is not specified
   */
  protected Class<?> getObjectClass()
    throws ClassNotFoundException
  {
    if ( hasAttribute( ATTR_OBJECT ))
    {
      String ostr = null;
      try {
        ostr = (String) getAttribute( ATTR_OBJECT );
        ostr = ostr.trim();
      }
      catch( MetaAttributeNotFoundException e ) {
        throw new RuntimeException( "Attribute was found but could not get it on MetaClass [" + this + "] and attribute [" + ATTR_OBJECT + "]" );
      }

      try {
        return Class.forName( ostr );
      }
      catch( ClassNotFoundException e ) {
        throw new ClassNotFoundException( "Specified Object Class [" + ostr + "] was not found", e );
      }
    }
    else
    {
      String ostr = getName().replaceAll( SEPARATOR, "." );

      try {
        return Class.forName( ostr );
      }
      catch( ClassNotFoundException e ) {
        log.debug( "Derived Object Class [" + ostr + "] was not found" );
      }
    }

    return null;
  }

  /**
   * Retrieves the attribute options expected for this class
   */
  public Collection<AttributeOption> getAttributeOptions()
  {
    ArrayList<AttributeOption> al = new ArrayList<AttributeOption>();
    al.add( new AttributeOption( ATTR_OBJECT, String.class, true, "The object class to instantiate" ));
    return al;
  }

  /**
   * Whether the MetaClass produces the object specified
   */
  public boolean produces( Object obj )
  {
    try {
      Class<?> cl = getObjectClass();
      if ( cl == null ) return false;

      boolean rc = cl.isInstance( obj );
      if ( rc ) return true;
    }
    catch( ClassNotFoundException e ) {};

    return false;
  }

  /**
   * Attaches a MetaClass to an Object
   * @param o Object to attach
   */
  public void attachMetaClass( Object o )
  {
    if ( o instanceof MetaClassAware )
    ((MetaClassAware) o ).setMetaClass( this );
  }

  /**
   * Sets the default values on an object
   * @param o Object to set the default values on
   */
  public void setDefaultValues( Object o )
    //throws MetaException
  {
    // Set the default values
    for( MetaField f : getMetaFields() )
    {
      Object val  = f.getDefaultValue();
      if ( val  != null )
        f.setObject( o, val );
    }
  }

  /**
   * Return a new MetaObject instance from the MetaClass
   */
  public Object newInstance()
    //throws MetaException
  {
    Class<?> oc = null;

    try {
      oc = getObjectClass();
      if ( oc == null )
        throw new MetaDataException( "Not Object Class was found" );
    }
    catch( ClassNotFoundException e ) {
      throw new MetaDataException( "Could find Object Class for MetaClass [" + toString() + "]", e );
    }

    try {
      if ( oc.isInterface() )
        throw new IllegalArgumentException( "Can not instantiate an Interface for MetaClass [" + toString() + "]" );

      Object o = oc.newInstance();

      // Attach the MetaClass
      attachMetaClass( o );

      // Set the Default Values
      setDefaultValues( o );

      return o;
    }
    catch( InstantiationException e ) {
      throw new RuntimeException( "Could not instantiate a new Object of Class [" + oc + "] for MetaClass [" + toString() + "]", e );
    }
    catch( IllegalAccessException e ) {
      throw new RuntimeException( "Illegal Access Exception instantiating a new Object for MetaClass [" + toString() + "]", e );
    }
  }

  ////////////////////////////////////////////////////
  // ABSTRACT METHODS

  /**
   * Retrieves the value from the object
   */
  public abstract Object getValue( MetaField f, Object obj ); //throws MetaException;

  /**
   * Sets the value on the object
   */
  public abstract void setValue( MetaField f, Object obj, Object val ); //throws MetaException;


  ////////////////////////////////////////////////////
  // MISC METHODS

  public Object clone()
  {
    MetaClass mc = (MetaClass) super.clone();
    //mc.mPackage = mPackage;
    return mc;
  }

    /* public String getDirtyFieldName() {
        return dirtyFieldName;
    }

    public void setDirtyFieldName(String dirtyFieldName) {
        this.dirtyFieldName = dirtyFieldName;
    }*/
}
