/*
 * Copyright 2003 Draagon Software LLC. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software LLC.
 * Use is subject to license terms.
 */

package com.draagon.meta.loader;

import com.draagon.meta.*;
import com.draagon.meta.attr.StringAttribute;
//import com.draagon.util.InitializationException;
import com.draagon.util.xml.XMLFileReader;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import java.io.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.CharacterData;
import org.w3c.dom.NodeList;

//import org.xml.sax.InputSource;
//import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;

/**
 * Meta Class loader for XML files
 */
public class XMLFileMetaClassLoader extends MetaClassLoader
{
	private static final long serialVersionUID = 6952160679341572048L;

	private static Log log = LogFactory.getLog( XMLFileMetaClassLoader.class );

    private boolean mTypesLoaded = false;

    private Hashtable<String,Class<?>> mClassTypes     = new Hashtable<String,Class<?>>();
    private Hashtable<String,Class<?>> mFieldTypes     = new Hashtable<String,Class<?>>();
    private Hashtable<String,Class<?>> mViewTypes      = new Hashtable<String,Class<?>>();
    private Hashtable<String,Class<?>> mValidatorTypes = new Hashtable<String,Class<?>>();
    private Hashtable<String,Class<?>> mAttrTypes      = new Hashtable<String,Class<?>>();

    private List<String> mSources = null;
    private String mTypes = null;
    
    public XMLFileMetaClassLoader()
    {
    }
    
    public void setSources( List<String> sources ) {
    	mSources = sources;
    }
    
    public List<String> getSources() {
    	return mSources;
    }
    
    public void setTypes( String types ) {
    	mTypes = types;
    }
    
    public String getTypes() { 
    	return mTypes;
    }

    @Override
    public void init() throws Exception {
    	
      if ( getSources() == null ) throw new IllegalStateException( "No Metadata Sources defined" );
    	
      super.init();
    	
      try {
        if ( getTypes() != null )
          loadTypes( getTypes() );
      }
      catch( MetaException e ) {
        log.error( "Could not load metadata types [" + getTypes() + "]: " + e.getMessage() );
        throw new IllegalStateException( "Could not load metadata types [" + getTypes() + "]", e );
      }

      // Load all the Meta sources
      for( Iterator<String> i = mSources.iterator(); i.hasNext(); )
      {
        String source = i.next();

        try {
          loadFromFile( source );
        }
        catch( MetaException e )
        {
          log.error( "Initialization of MetaClassLoader [" + source + "] failed: " + e.getMessage() );
          throw new IllegalStateException( "Initialization of MetaClassLoader [" + source + "] failed", e );
        }
      }    	
    }
    
    /**
     * Loads the specified resource
     */
    public void loadTypes( String file )
        throws MetaException
    {
      // LOAD THE TYPES XML FILE
      if ( file == null )
          throw new MetaException( "The Types XML file was not specified" );

      InputStream is = null;

      File f = new File( file );
      if ( f.exists() )
      {
        try {
          is = new FileInputStream( f );
        }
        catch( Exception e ) {
          log.error( "Can not read Types XML file [" + file + "]: " + e.getMessage() );
          throw new MetaException( "Can not read Types XML file [" + file + "]: " + e.getMessage(), e );
        }
      }
      else
      {
        is = getClass().getClassLoader().getResourceAsStream( file );
        if ( is == null )
        {
          log.error( "Types XML file [" + file + "] does not exist" );
          throw new MetaException( "The Types XML item file [" + file + "] was not found" );
        }
      }

      try
      {
        loadTypesFromStream( is );
      }
      catch( MetaException e )
      {
        log.error( "Meta Types XML [" + file + "]: " + e.getMessage() );
        throw new MetaException( "The Types XML file [ " + file + "] could not be loaded: " + e.getMessage(), e );
      }

      mTypesLoaded = true;
    }

    /**
     * Loads all the classes specified in the Filename
     */
    public void loadTypesFromStream( InputStream is )
        throws MetaException
    {
      Document doc = null;

      try {
        doc = XMLFileReader.loadFromStream( is );
      }
      catch( IOException e ) {
        log.error( "IO Error loading Types XML: " + e.getMessage() );
        throw new MetaException( "IO Error loading Types XML: " + e.getMessage(), e );
      }
      finally {
        try { is.close(); } catch( Exception e ) {}
      }

      //////////////////////////////////////////////////////
      // PARSE THE TYPES XML

      try
      {
          // Look for the <items> element
          Collection<Element> elements = getElementsOfName( doc, "types" );
          if ( elements.size() == 0 )
              throw new MetaException( "The root 'types' element was not found" );

          Element el = elements.iterator().next();

          loadAllTypes( el );
      }
      catch( SAXException e )
      {
          throw new MetaException( "Parse error loading MetaClasses: " + e.getMessage(), e );
      }
    }

    /**
     * Loads the specified group types
     */
    protected void loadAllTypes( Element el )
      throws MetaException, SAXException
    {
      loadTypes( el, "class", mClassTypes );
      loadTypes( el, "field", mFieldTypes );
      loadTypes( el, "view", mViewTypes );
      loadTypes( el, "validator", mValidatorTypes );
      loadTypes( el, "attr", mAttrTypes );
    }

    /**
     * Loads the specified group types
     */
    protected void loadTypes( Element root, String section, Hashtable<String,Class<?>> map )
      throws MetaException, SAXException
    {
      Collection<Element> c = getElementsOfName( root, section );

      // Iterate through each section grouping (should just be 1)
      for( Element el : c )
      {
        Collection<Element> typeCol = getElementsOfName( el, "type" );

        // Iterate through each type
        for( Element typeEl : typeCol )
        {
          String name = typeEl.getAttribute( "name" );
          String tclass = typeEl.getAttribute( "class" );

          if ( name.length() == 0 )
            throw new MetaException( "Type of section [" + section + "] has no 'name' attribute specified" );

          try {
            Class<?> tcl = Class.forName( tclass );

            // Add the type class with the specified name
            map.put( name, tcl );
          }
          catch( ClassNotFoundException e ) {
             throw new MetaException( "Type of section [" + section + "] with name [" + name + "] has invalid class: " + e.getMessage() );
             //log.warn( "Type of section [" + section + "] with name [" + name + "] has unknown class: " + e.getMessage() );
          }
        }
      }
    }

    /**
     * Loads all the classes specified in the Filename
     */
    public void loadFromFile( String file )
        throws MetaException
    {
      // LOAD THE XML FILE
      if ( file == null )
        throw new MetaException( "The Meta XML file was not specified" );

      InputStream is = null;

      File f = new File( file );
      if ( f.exists() )
      {
          try {
            is = new FileInputStream( f );
          }
          catch( Exception e ) {
            log.error( "Can not read Meta XML file [" + file + "]: " + e.getMessage() );
            throw new MetaException( "Can not read Meta XML file [" + file + "]", e );
          }
      }
      else
      {
          is = getClass().getClassLoader().getResourceAsStream( file );
          if ( is == null )
          {
              log.error( "Meta XML file [" + file + "] does not exist" );
              throw new MetaException( "The Meta XML item file [" + file + "] was not found" );
          }
      }

      try
      {
          loadFromStream( is );
      }
      catch( MetaException e )
      {
        log.error( "Meta XML file [" + file + "]: " + e.getMessage() );
        throw new MetaException( "The Meta XML file [ " + file + "] could not be loaded", e );
      }
      catch( MetaDataException e )
      {
        log.error( "Meta XML file [" + file + "]: " + e.getMessage() );
        throw new MetaException( "The Meta XML file [ " + file + "] could not be loaded", e );
      }
    }

    /**
     * Loads all the classes specified in the Filename
     */
    public void loadFromStream( InputStream is )
        throws MetaException
    {
            Document doc = null;


            try {
              doc = XMLFileReader.loadFromStream( is );
            }
            catch( IOException e ) {
              log.error( "IO Error loading Meta XML: " + e.getMessage() );
              throw new MetaException( "IO Error loading Meta XML: " + e.getMessage(), e );
            }
            finally {
              try { is.close(); } catch( Exception e ) {}
            }

            ////////////////////////////////////////////////////////
            // LOAD DEFAULT TYPES

            // If no types found, then load the default types
            if ( !mTypesLoaded )
              loadTypes( "com/draagon/meta/loader/meta.types.xml" );

            //////////////////////////////////////////////////////
            // PARSE THE ITEMS XML BLOCK

            try
            {
                // Look for the <items> element
                Collection<Element> elements = getElementsOfName( doc, "metadata" );
                if ( elements.size() == 0 )
                    throw new MetaException( "The root 'meta' element was not found" );

                Element itemdocElement = elements.iterator().next();

                String defaultPackageName = itemdocElement.getAttribute( "package" );
                if ( defaultPackageName == null || defaultPackageName.trim().length() == 0 )
                  defaultPackageName = "";
                    //throw new MetaException( "The Meta XML had no 'package' attribute defined" );

                // Load the types
                String types = itemdocElement.getAttribute( "types" );
                if ( types != null && types.length() > 0 )
                  loadTypes( types );

                // Load any types specified in the Metadata XML
                try
                {
                    // Look for the <items> element
                    Collection<Element> typeElements = getElementsOfName( itemdocElement, "types" );
                    if ( typeElements.size() > 0 )
                    {
                      Element typeEl = typeElements.iterator().next();

                      String typeFile = typeEl.getAttribute( "file" );
                      if ( typeFile.length() > 0 )
                        loadTypes( typeFile );
                      else
                        loadAllTypes( typeEl ); // Load inner tags
                    }
                }
                catch( SAXException e )
                {
                    throw new MetaException( "Parse error loading MetaClasses: " + e.getMessage(), e );
                }

                parseClasses( defaultPackageName, itemdocElement );
            }
            catch( SAXException e )
            {
                throw new MetaException( "Parse error loading MetaClasses: " + e.getMessage(), e );
            }
        }

    protected void parseClasses( String defaultPackageName, Element element ) throws MetaException, SAXException
    {
        for( Element itemElement : getElementsOfNames( element, new String[] {"metaclass","class"} ))
        {
            // get the item name
            String name = itemElement.getAttribute( "name" );
            if ( name == null || name.equals( "" ))
            throw new MetaException( "MetaClass had no name specfied in XML" );

            String packageName = itemElement.getAttribute( "package" );
            if ( packageName == null || packageName.trim().length() == 0 )
              packageName = defaultPackageName;

                        // Load or get the MetaClass
            MetaClass mc = null;
            boolean isNewClass = false;

            try {
              if ( packageName.length() > 0 )
                mc = getMetaClass( packageName + MetaClass.SEPARATOR + name );
              else
                mc = getMetaClass( name );
            } catch( MetaClassNotFoundException e ) {}

            if ( mc == null )
            {
                isNewClass = true;

                // Set the SuperClass if one is defined
                MetaClass superClass = null;
                String superStr = itemElement.getAttribute( "super" );

                // If a super class was specified
                if ( superStr.length() > 0 )
                {
                  // Try to find it with the name prepended if not fully qualified
                  try {
                    if ( superStr.indexOf( MetaClass.SEPARATOR ) < 0 && packageName.length() > 0 )
                      superClass = getMetaClass( packageName + MetaClass.SEPARATOR + superStr );
                  }
                  catch( MetaClassNotFoundException e ) {
                    log.debug( "Could not find MetaClass [" + packageName + MetaClass.SEPARATOR + superStr + "], assuming fully qualified" );
                  }

                  if ( superClass == null )
                  {
                    try {
                        superClass = getMetaClass( superStr );
                    }
                    catch( MetaClassNotFoundException e ) {
                        log.error( "Invalid MetaClass [" + mc + "], the SuperClass [" + superStr + "] does not exist" );
                        throw new MetaException( "Invalid MetaClass [" + mc + "], the SuperClass [" + superStr + "] does not exist" );
                    }
                  }
                }

                // get the class reference and create the class
                String typeName = itemElement.getAttribute( "type" );

                Class<?> c = null;
                try {
                  // Attempt to load the referenced class
                  if ( typeName.equals( "" ))
                  {
                    if ( superClass != null )
                      c = superClass.getClass();
                    else
                      //c = BeanMetaClass.class;
                      throw new MetaException( "MetaClass [" + name + "] has no type defined" );
                  }
                  else
                  {
                    c = (Class<?>) mClassTypes.get( typeName );
                    if ( c == null )
                      throw new MetaException( "MetaClass type [" + typeName + "] was not recognized" );
                  }

                  mc = (MetaClass) c.newInstance();
                }
                catch( MetaException e ) {
                  throw e;
                }
                catch( Exception e ) {
                  log.error( "Invalid MetaClass [" + c.getName() + "]: " + e.getMessage() );
                  throw new MetaException( "Invalid MetaClass [" + c.getName() + "]", e );
                }

              // Set the name and package name
              mc.setName( packageName + MetaClass.SEPARATOR + name );
              //mc.setPackage( packageName );

              mc.setSuperClass( superClass );
            }

            // Parse and set the Attributes
            Collection<MetaAttribute> attrs = parseAttributes( itemElement );
            for ( Iterator<MetaAttribute> j = attrs.iterator(); j.hasNext(); ) {
                mc.addAttribute( j.next() );
            }

            // Parse the fields
            parseFields( mc, itemElement );

            // Get the default view settings
            Collection<MetaView> def_views = parseViews( name, itemElement, null );

            // If not views were defined, then use the default views
            if ( def_views.size() > 0 )
            {
                for( MetaField f : mc.getMetaFields() )
                {
                    for( MetaView v : def_views )
                    {
                        if ( !f.hasView( v.getName() ))
                            f.addMetaView( (MetaView) v.clone() );
                    }
                }
            }

            // Add the MetaClass to the loader
            if ( isNewClass ) addMetaClass( mc );
          }
    }

  protected void parseFields( MetaClass mc, Element e ) throws MetaException, SAXException
  {
    for( Element fieldElement : getElementsOfNames( e, new String[] {"metafield","field"} ))
    {
      String name = fieldElement.getAttribute( "name" );
      String type = fieldElement.getAttribute( "type" );
      String def = fieldElement.getAttribute( "default" );

      //int type = 0;

      // Get the field name
      if ( name == null || name.equals( "" ))
        throw new MetaException( "Field in MetaClass [" + mc + "] had no name" );

      MetaField field = null;
      boolean isNewField = false;

      Class<?> classType = null;

      // If the type is defined, then load the class
      if ( !type.equals( "" ))
      {
        classType = mFieldTypes.get( type );
        if ( classType == null )
          throw new MetaException( "MetaField [" + name + "] has type [" + type+ "] that is not recognized" );
      }

      // If a field with that name exists, then load it
      if ( mc.hasMetaField( name )) {
        field = mc.getMetaField( name );
      }

      // If there is no type or validator, then throw an exception
      if ( classType == null && field == null )
          throw new MetaException( "MetaField [" + name + "] of MetaClass [" + mc + "] had no type" );

      // If the field exists and is of the same class, then wrap it
      if ( field != null &&
          ( field.getClass() == classType || classType == null ))
      {
        if ( field.getDeclaringMetaClass() != mc )
        {
          field = (MetaField) field.wrap();
          isNewField = true;
        }
      }

      // Otherwise, create a new field
      else
      {
        if ( field != null )
          throw new MetaException( "MetaField [" + name + "] of MetaClass [" + mc + "] is already defined as type [" + field.getClass() + "]" );

        try {
          field = (MetaField) classType.newInstance();
          field.setName( name );
          isNewField = true;
        }
        catch (InstantiationException ex) {
          throw new RuntimeException( "Instantiation exception creating new MetaField of Class [" + classType + "]", ex );
        }
        catch (IllegalAccessException ex) {
          throw new RuntimeException( "Illegal access exception creating new MetaField of Class [" + classType + "]", ex );
        }
      }


      // Parse and set the Attributes
      Collection<MetaAttribute> attrs = parseAttributes( fieldElement );
      for ( Iterator<MetaAttribute> a = attrs.iterator(); a.hasNext(); ) {
          field.addAttribute( a.next() );
      }

      // Parse the Views
      for ( MetaView tv : parseViews( name + "@" + mc.getName(), fieldElement, field ))
      {
          if ( !field.hasView( tv.getName() ) || tv.getDeclaringMetaField() != field )
              field.addMetaView( tv );

          //ystem.out.println( "VIEW: " + tv );
      }

      // Add the Validators
      for ( MetaValidator tv : parseValidators( name + "@" + mc.getName(), fieldElement, field ) ) 
      {
          if ( !field.hasValidator( tv.getName() ))
              field.addMetaValidator( tv );
      }

      // Set the default value
      if ( def != null && def.length() > 0 ) {
          field.setDefaultValue( def );
      }

      // If it's a new field, then add it
      if ( isNewField ) {
          mc.addMetaField( field );
      }
    }
  }

  protected Collection<MetaView> parseViews( String parent, Element e, MetaField mf )
    throws MetaException, SAXException
  {
    try {
        Class.forName( "com.draagon.web.meta.view.WebView" );
    }
    catch( ClassNotFoundException cnfe) {
        // The meta-web package is missing, so assume EJB server mode
        log.debug( "No WebView found, assuming server-side installation" );

        return new ArrayList<MetaView>();
    }

    ArrayList<MetaView> vv = new ArrayList<MetaView>();

    for( Element viewElement : getElementsOfNames( e, new String[] {"metaview","view"} )) 
    {
      String vname = viewElement.getAttribute( "name" );
      String vtype = viewElement.getAttribute( "type" );

      if ( vname.equals( "" ))
        throw new MetaException( "View Element [" + e.getNodeName() + "] had no name defined" );

      if ( vtype.equals( "" ))
        throw new MetaException( "MetaView [" + vname + "] of [" + parent + "] had no type defined" );

      MetaView view = null;
      boolean isNewView = false;

      Class<?> classType = null;

      // If the type is defined, then load the class
      if ( !vtype.equals( "" ))
      {
        classType = mViewTypes.get( vtype );
        if ( classType == null )
        {
          log.warn( "MetaView [" + vname + "] of [" + parent + "] has type [" + vtype+ "] that is not recognized" );
          //throw new MetaException( "MetaValidator [" + name + "] of [" + parent + "] has type [" + type+ "] that is not recognized" );
          continue;
        }
      }

      // If a view with that name exists, then load it
      if ( mf != null && mf.hasView( vname ) && classType == null )
        view = mf.getView( vname );

      // If there is no type or validator, then throw an exception
      if ( classType == null && view == null )
          throw new MetaException( "View [" + vname + "] MetaField [" + mf + "] had no type" );

      // If the validator exists and is of the same class, then wrap it
      if ( view != null &&
          ( view.getClass() == classType || classType == null ))
      {
        if ( view.getDeclaringMetaField() != mf )
        {
          view = (MetaView) view.wrap();
          isNewView = true;
        }
      }

      // Otherwise, create a new view
      else
      {
        try {
          view = (MetaView) classType.newInstance();
          view.setName( vname );
          isNewView = true;
        }
        catch (InstantiationException ex) {
          throw new RuntimeException( "Instantiation exception creating new MetaView of Class [" + classType + "]", ex );
        }
        catch (IllegalAccessException ex) {
          throw new RuntimeException( "Illegal access exception creating new MetaView of Class [" + classType + "]", ex );
        }
      }

      Collection<MetaAttribute> attrs = parseAttributes( viewElement );

      for ( Iterator<MetaAttribute> j = attrs.iterator(); j.hasNext(); ) {
        view.addAttribute( j.next() );
      }

      if ( isNewView ) {
        vv.add( view );
      }
    }

    return vv;
  }

  protected Collection<MetaValidator> parseValidators( String parent, Element e, MetaField mf )
    throws MetaException, SAXException
  {
    ArrayList<MetaValidator> vv = new ArrayList<MetaValidator>();

    for( Element el : getElementsOfNames( e, new String[] {"metavalidator","validator"} ))
    {
      String name = el.getAttribute( "name" );
      String type = el.getAttribute( "type" );

      if ( name.equals( "" ))
        throw new MetaException( "Validator Element [" + e.getNodeName() + "] had no name defined" );

      if ( type.equals( "" ))
        throw new MetaException( "MetaValidator [" + name + "] of [" + parent + "] had no type defined" );

      Collection<MetaAttribute> attrs = parseAttributes( el );

      MetaValidator validator = null;
      boolean isNewValidator = false;

      Class<?> classType = null;

      // If the type is defined, then load the class
      if ( !type.equals( "" ))
      {
        classType = mValidatorTypes.get( type );
        if ( classType == null )
        {
          log.warn( "MetaValidator [" + name + "] of [" + parent + "] has type [" + type+ "] that is not recognized" );
          //throw new MetaException( "MetaValidator [" + name + "] of [" + parent + "] has type [" + type+ "] that is not recognized" );
          continue;
        }
      }

      // If a validator with that name exists, then load it
      if ( mf.hasValidator( name ))
        validator = mf.getValidator( name );

      // If there is no type or validator, then throw an exception
      if ( classType == null && validator == null )
          throw new MetaException( "Validator [" + name + "] MetaField [" + mf + "] had no type" );

      // If the validator exists and is of the same class, then wrap it
      if ( validator != null &&
          ( validator.getClass() == classType || classType == null ))
      {
        if ( validator.getDeclaringMetaField() != mf )
        {
          validator = (MetaValidator) validator.wrap();
          isNewValidator = true;
        }
      }

      // Otherwise, create a new validator
      else
      {
        try {
          validator = (MetaValidator) classType.newInstance();
          validator.setName( name );
          isNewValidator = true;
        }
        catch (InstantiationException ex) {
          throw new RuntimeException( "Instantiation exception creating new MetaValidator of Class [" + classType + "]", ex );
        }
        catch (IllegalAccessException ex) {
          throw new RuntimeException( "Illegal access exception creating new MetaValidator of Class [" + classType + "]", ex );
        }
      }

      // Set the attributes
      for ( Iterator<MetaAttribute> j = attrs.iterator(); j.hasNext(); ) {
          validator.addAttribute( j.next() );
      }

      if ( isNewValidator )
        vv.add( validator );
    }

    return vv;
  }

  protected Collection<MetaAttribute> parseAttributes( Element e ) throws MetaException, SAXException
  {
    ArrayList<MetaAttribute> attrs = new ArrayList<MetaAttribute>();

    for( Element el : getElementsOfName( e, "attr" ))
    {
      // Get the attribute name
      String name = el.getAttribute( "name" );
      String type = el.getAttribute( "type" );

      if ( name.equals( "" ))
        throw new MetaException( "Attribute Element [" + e.getNodeName() + "] had no name defined" );

      Class<?> c = null;
      if ( type.equals( "" ))
        c = StringAttribute.class;
      else
      {
        c = mAttrTypes.get( type );
        if ( c == null )
          throw new MetaException( "MetaAttribute [" + name + "] has type [" + type+ "] that is not recognized" );
      }

      MetaAttribute attr = null;
      try
      {
        attr = (MetaAttribute) c.newInstance();
        attr.setName( name );
      }
      catch( Exception ex ) {
        log.error( "Invalid MetaAttribute [" + name + "]: " + ex.getMessage() );
        throw new MetaException( "Invalid MetaAttribute [" + name + "]", ex );
      }

      ///////////////////////////////////////////////////
      // Get the Node value

      // Get the first node
      Node nv = el.getFirstChild();

      // Loop through and ignore the comments
      while ( nv != null && nv.getNodeType() == Node.COMMENT_NODE ) nv.getNextSibling();

      // If a valid node exists, then get the data
      if ( nv != null ) {
                switch( nv.getNodeType() )
                {
                    // If CDATA just set the whole thing
                    case Node.CDATA_SECTION_NODE:
                        attr.setValue( ((CharacterData) nv).getData() );
                        break;

                    // If an Element just pass it in for parsing
                    case Node.ELEMENT_NODE:
                        attr.setValue( nv );
                        break;

                    // If just text, then pass it in
                    case Node.TEXT_NODE:
                        attr.setValue( nv.getNodeValue() );
                        break;

                    default:
                        log.warn( "Unsupported Node Type for node [" + nv + "]" );
                }
            }

            // Add the attribute
      attrs.add( attr );
    }

    return attrs;
  }

    /**
     * Returns a collection of child elements of the given name
     */
  protected Collection<Element> getElementsOfName( Node n, String name )
  {
        ArrayList<Element> elements = new ArrayList<Element>();
        if ( n == null ) return elements;

        NodeList list = n.getChildNodes();
        for( int i = 0; i < list.getLength(); i++ )
        {
            Node node = list.item( i );
            if ( node instanceof Element
                    && node.getNodeName().equals( name ))
            {
                elements.add( (Element) node );
            }
        }

        return elements;
    }

    /**
     * Returns a collection of child elements of the given name
     */
  protected Collection<Element> getElementsOfNames( Node n, String [] names )
  {
        ArrayList<Element> elements = new ArrayList<Element>();
        if ( n == null ) return elements;

        NodeList list = n.getChildNodes();
        for( int i = 0; i < list.getLength(); i++ )
        {
            Node node = list.item( i );
            if ( node instanceof Element )
            {
                for( int j = 0; j < names.length; j++ )
                {
                    if ( node.getNodeName().equals( names[ j ] ))
                        elements.add( (Element) node );
                }
            }
        }

        return elements;
    }
}
